#ifndef _LGSDIREKT_H_
#define _LGSDIREKT_H_

// Hilfsprogramme 
#include <iostream>
using namespace std;
#include <cstdlib>

void TransformCSRtoDense(int* column_indices,double* entries,
			 int* row_offsets,int n,double* a)
// Transformiert Matrix aus CSR-Format in volles Format
// als 1D array gespeichert		      
{
  int k1,k2,k,j,i,nnz,m;
  for  (i=0;i<n;i++)
    {
      k1=row_offsets[i];
      k2=row_offsets[i+1];
      for (k=k1;k<k2;k++)
	{
           j=column_indices[k];
           m=i*n+j;
           a[m]=entries[k];
	}
    }
}


void pivotsearch(double* a,int n,int *j,int i)
// Pivotsuche in iter Spalte und Tausch der Zeilen j und i
{
  double* h;
  int k,k1,k2,l;
  double value;
  h=(double*)malloc(n*sizeof(double));
  k=(n+1)*i;
  vdcopy(n-i,h,1,&a[k],n);
  vdbmax(h,n-i,&value,&l);
  *j=l+i;
  // Tausche jmax mit iter Zeile
  k1=i*n;
  k2=(*j)*n;
  vdcopy(n,h,1,&a[k1],1);
  vdcopy(n,&a[k1],1,&a[k2],1);
  vdcopy(n,&a[k2],1,h,1);
  free(h);
}



void LUDecompDense(double* a,int n,int pivot,int* P)
// LU-Zerlegung fuer Matrix A\in R^{n,n}
//  als 1D array zeilenweise gespeichert
// Falls pivot>0, dann Spaltenpivotsuche
// P zeigt die permuationen an
// ki und kj zeigen jeweils den Beginn der i-ten und j-ten Zeile in A an
// ab a_ii bzw. a_ij
{
  int i,j,kj,ki,jmax;
  double value,aii,aij,lij;
  ki=0;
  for (i=0;i<n;i++)
    {
      P[i]=i;
    }
  for (i=0;i<n-1;i++) 
    {
      if (pivot>0)
	{
	  pivotsearch(a,n,&jmax,i);  
          if (i!=jmax)
	    {        
	      Change(&P[i],&P[jmax]);
            }
	}
      kj=ki+n;
      for (j=i+1;j<n;j++)
	{
          aii=a[ki];
          aij=a[kj];
          lij=-aij/aii;
          a[kj]=-lij;
          vdaxpy(n-i-1,&a[ki+1],lij,&a[kj+1],&a[kj+1]); 
	  kj=kj+n;
        }
      ki=ki+n+1;  
    }
}

void  VorEliDense(double* a,double* b,int n)
// Vorwaertselimination, d.h. Loese L c=b 
// A=LU als 1D array gespeichert
{
  int i,k;
  double sigma;
  for (i=1;i<n;i++) 
    {
      k=i*n;
      dscapr(&sigma,i,b,&a[k]);
      b[i]=b[i]-sigma;
    } 
} 

void  RueckEliDense(double* a,double* b,int n)
// Vorwaertselimination, d.h. Loese U c=b 
// A=LU als 1D array gespeichert
{
  int i,k;
  double sigma;
  for (i=n-1;i>-1;i--) 
    {
      k=i*(n+1);
      dscapr(&sigma,n-1-i,&b[i+1],&a[k+1]);
      b[i]=(b[i]-sigma)/a[k];
     } 
} 

void vcopyperm(int n,int* P,double* x,double* b)
{
  int i;
  for (i=0;i<n;i++) 
    {
      x[i]=b[P[i]];
    }
}

void GaussDense(double* a,double* b,double* x,int n,int pivot)
// Loese Ax=b mittels Gauss (mit/ohne Pivotsierung)
{
  int* P;
  P=(int*)malloc(n*sizeof(int));
  LUDecompDense(a,n,pivot,P);
  if (pivot>0)
    {
      vcopyperm(n,P,x,b);
    }
  else
    {
      vdcopy(n,x,1,b,1);
    }
  VorEliDense(a,x,n);
  RueckEliDense(a,x,n);
}

void LUDecompVBZ(int* column_indices,double* entries,int* row_offsets,
		 int* row_indices,int n)
// LU-Zerlegung fuer Matrix A\in R^{n,n}
//  als VBZ-Matrix gespeichert
// ohne Pivotisierung
// !!! Achtung: Besetzung der Matrix sollte symmetrisch besetzt sein
{
  int i,j,jmin,jmax,k1,k2,ijmin,ijmax,j1,j2;  
  double aii,aij,lij;
  int iimin,iimax,imin,imax;
  k1=row_offsets[0];
 
  for (i=0;i<n-1;i++) 
  {    
    k2=row_offsets[i+1]; // k2-k1 Bandbreite
   // Bereich ite Zeile in entries von k1 bis k2-1
   // Eliminiere nun die Eintraege Spalte i von jmin bis jmax-1
   jmin=i+1; 
   jmax=row_indices[i]+1; //jmax<=n+1
   // Bereich der NNE-Eintraege in Zeile i
   iimin=jmin;
   iimax=column_indices[i]+row_offsets[i+1]-row_offsets[i];
   // Hauptdiagonalelement
   j2=k1+i-column_indices[i];
   aii=entries[j2];
   for (j=jmin;j<jmax;j++)
     { 
       // Bereich jte Zeile mit NNE
       ijmin=column_indices[j];
       ijmax=ijmin+row_offsets[j+1]-row_offsets[j]; // ijmax<=n
       if (ijmin<i+1) // sonst nichts zu eliminieren
	 {
	   // Beginn der Eintraege zum Eliminieren
      	  j1=row_offsets[j]+i-column_indices[j];
      	  aij=entries[j1];
          lij=-aij/aii;
          entries[j1]=-lij;
          // Spaltenbereich bei elimination a_{ij}
          imax=min(iimax,ijmax);
          imin=max(ijmin,iimin);
          vdaxpy(imax-imin,&entries[j2+1],lij,&entries[j1+1],&entries[j1+1]);
          }
     }
   k1=k2;
  }
}


void VorEliVBZ(int* column_indices,double* entries,int* row_offsets,
	       int* row_indices,double* b,int n)
// Vorwaertselimination einer Matrix A=LU \in R^{n,n} in VBZ-Format
// d.h. Loese L c= b, Speicherung in b.
{
  int i,k1,k2,j1,j2;
  double sigma;
  for (i=1;i<n;i++) 
    {
      // j1..j2-1 Bereich der Eliminierenden Spalte
      // k1..k2-1 Bereich der Eintraege
      j1=column_indices[i];
      j2=i;
      k1=row_offsets[i];
      k2=k1+j2-j1;
      dscapr(&sigma,j2-j1,&b[j1],&entries[k1]);
      b[i]=b[i]-sigma;
    } 
}

void RueckEliVBZ(int* column_indices,double* entries,int* row_offsets,
	       int* row_indices,double* b,int n)
// Rueckwaertselimination einer Matrix A=LU \in R^{n,n} in VBZ-Format
// d.h. Loese U c= b, Speicherung in b.
{
  int i,k1,k2,j1,j2;
  double sigma;
  for (i=n-1;i>-1;i--) 
    {
      // j1..j2-1 Bereich der Eliminierenden Spalte
      // k1..k2-1 Bereich der Eintraege
      j1=i+1;
      k1=j1+row_offsets[i]-column_indices[i];
      k2=row_offsets[i+1];
      j2=k2+j1-k1; 
      dscapr(&sigma,j2-j1,&b[j1],&entries[k1]);
      b[i]=(b[i]-sigma)/entries[k1-1];
    } 
}

void GaussVBZ(int* column_indices,double* entries,int* row_offsets,
	      int* row_indices,double* b,double* x,int n)
// Loest das LGS Ax=b mittels Gauss (ohne Pivotisierung)
// Matrix A=LU \in R^{n,n} in VBZ-Format gespeichert
{
  vdcopy(n,x,1,b,1);
  LUDecompVBZ(column_indices,entries,row_offsets,row_indices,n);
  VorEliVBZ(column_indices,entries,row_offsets,row_indices,x,n);
  RueckEliVBZ(column_indices,entries,row_offsets,row_indices,x,n);
}

#endif
